/*
  ExeHack 1.0
  Loader for scripting patches with lua
  Copyright 2006 Bryan Steinbach
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * 1.0:
 * Added sleep loop to ensure process initialized before dll injection
 * switched to luaL_openlib
 *
 * 0.4:
 * Moved all hacks out to lua
 *
 * 0.3a:
 * fixed command line bungling that made 0.3 nigh useless
 *
 * 0.3:
 * allows one to always enable disabled triggers
 * removes nag dialog when saving with default map description
 * get rid of prepend with war3mapImported on imported files
 * execute tmap.bat instead of war3.exe by default
 * disable syntax checking
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <direct.h>

extern "C" {
#include "../lua/lua.h"
#include "../lua/lauxlib.h"
#include "../lua/lualib.h"
}

//both getopts remove argument from command line

//tells you if it's present
int getopt_noarg(int &argc, char **argv, char *optname)
{
	for(int i=0;i<argc;i++) {
		if(strcmp(optname,argv[i]) == 0) {
			argc--;
			for(int j=i;j<argc;j++) {
				argv[j] = argv[j+1];
			}
			return true;
		}
	}
	return false;
}

//Sends the next arg on the line back
char *getopt_arg(int &argc, char **argv, char *optname)
{
	for(int i=0;i<argc;i++) {
		if(strcmp(optname,argv[i]) == 0) {
			if (i+1 < argc) {
				argc-=2;
				char *arg = argv[i+1];
				for(int j=i;j<argc;j++) {
					argv[j] = argv[j+2];
					argv[j+2] = argv[j+3];
					
				}
				return arg;
			}
		}
	}
	return NULL;
}
void report_errors(lua_State *lua, int status)
{
	if ( status!=0 ) {
		fprintf(stderr,"-- %s\n", lua_tostring(lua,-1));
		lua_pop(lua, 1); // remove error message
	}
}

void inject(HANDLE hProcess, int address, void *newcode, int len)
{
	DWORD old,old2;
	int ret = VirtualProtectEx(hProcess,(void *)address,len,PAGE_EXECUTE_READWRITE,&old);
	ret |= WriteProcessMemory(hProcess,(void *)address,newcode,len,NULL);
	if(ret) {
		printf("Successfully patched %x with %d bytes\n",address,len);
	}
	else {
		printf("Could not patch %x with %d bytes\n",address,len);
	}
	VirtualProtectEx(hProcess,(void *)address,len,old,&old2);
}

int lua_patchdword(lua_State *l)
{
	HANDLE proc = (HANDLE)lua_tointeger(l,1);
	int addr = lua_tointeger(l,2);
	int dword = lua_tointeger(l,3);
	inject(proc,addr,&dword,4);
	lua_pushinteger(l,0);
	return 1;
}

int lua_patchbyte(lua_State *l)
{
	HANDLE proc = (HANDLE)lua_tointeger(l,1);
	int addr = lua_tointeger(l,2);
	char byte = lua_tointeger(l,3);
	inject(proc,addr,&byte,1);
	lua_pushinteger(l,0);
	return 1;
}

int lua_sleep(lua_State *l)
{
	int ms = lua_tointeger(l,1);
	Sleep(ms);
	lua_pushinteger(l,0);
	return 1;
}

int lua_patchstr(lua_State *l)
{
	HANDLE proc = (HANDLE)lua_tointeger(l,1);
	int addr = lua_tointeger(l,2);
	char *str = (char *)lua_tostring(l,3);
	inject(proc,addr,str,strlen(str));
	lua_pushinteger(l,0);
	return 1;
}

void nopout(HANDLE hProcess, int addr, int len)
{
	int failed = 0;
	char nop = '\x90';
	int end = addr+len;
	for(int i=addr;i<end;i++) {
		failed |= !WriteProcessMemory(hProcess,(void *)i,&nop,1,NULL);
	}
	if(failed)
		printf("could not wipe %x with %d nops\n",addr,len);
	else
		printf("Successfully wiped %x with %d nops\n",addr,len);
}

int lua_nopout(lua_State *l)
{
	HANDLE proc = (HANDLE) lua_tointeger(l,1);
	int addr = lua_tointeger(l,2);
	int len = lua_tointeger(l,3);
	nopout(proc,addr,len);
	lua_pushinteger(l,0);
	return 1;
}

void xoreax(HANDLE proc, int address)
{
	char op[2] = { 0x31, 0xC0 };
	inject(proc,address,op,2);
}

int lua_xoreax(lua_State *l)
{
	HANDLE proc = (HANDLE)lua_tointeger(l,1);
	int address = lua_tointeger(l,2);
	xoreax(proc,address);
	lua_pushinteger(l,0);
	return 1;
}

HANDLE startexe(char *exefile,char *exestring)
{
//	printf("%s\t%s\n",exefile,exestring);
	STARTUPINFO si;
	ZeroMemory(&si,sizeof(si)); // Clean
	si.cb = sizeof(si); // Set size
	PROCESS_INFORMATION pi; // Info
	ZeroMemory(&pi,sizeof(pi)); // Clean
	//CreateProcess(exefile,exestring,NULL,NULL,false,CREATE_SUSPENDED,NULL,NULL,&si,&pi);
	CreateProcess(exefile,exestring,NULL,NULL,false,0,NULL,NULL,&si,&pi);
	if(!pi.hProcess) {
		printf("Could not start %s\n",exefile);
	}
	ResumeThread(pi.hThread);
	return pi.hProcess;
}

int lua_startexe(lua_State *l)
{
	char *exefile = (char *)lua_tostring(l,1);
	char *exestring = (char *)lua_tostring(l,2);
	lua_pushinteger(l,(int)startexe(exefile,exestring));
	return 1;
}

int scanbuf(char *buf, int blen, char *pattern, int plen)
{
	for(int i=0;i<blen-plen+1;i++) {
		if(memcmp(buf,pattern,plen) == 0) return i;
	}
	return -1;
}

//Won't catch matches across boundaries of regions
unsigned long scanmem(HANDLE proc, char *pattern, int length)
{
	SYSTEM_INFO si;
	MEMORY_BASIC_INFORMATION mbi;
	unsigned long i;
	unsigned long ret;
	try {
		GetSystemInfo(&si);
		i=(unsigned long)si.lpMinimumApplicationAddress;
		for(;i<(unsigned long)si.lpMaximumApplicationAddress;i++) {
			mbi.RegionSize = 0;
			ret = VirtualQueryEx(proc,(void *)i,&mbi,sizeof(mbi));
			if(ret == sizeof(mbi) && mbi.Type == MEM_PRIVATE 
					&& mbi.State == MEM_COMMIT && mbi.RegionSize > 0 ) {
				char *buf = (char *)malloc(mbi.RegionSize);
				if(buf == NULL) throw((char *)"couldn't allocate buf");
				unsigned long written;
				ret = ReadProcessMemory(proc,mbi.BaseAddress,(void *)buf,mbi.RegionSize,&written);
				if(!ret || written != mbi.RegionSize) throw((char *)"couldn't read process memory");
				
				int offset = scanbuf(buf,mbi.RegionSize,pattern,length);
				if(offset >= 0) return offset+(unsigned long)mbi.BaseAddress;
				free(buf);
			}
		}
	} catch(const char *errormsg) {
		fprintf(stderr,"%s\n",errormsg);
		return 0;
	}

	return 0;
}

int lua_findstring(lua_State *l)
{
	HANDLE proc = (HANDLE)lua_tointeger(l,1);
	char *pattern = (char *)lua_tostring(l,2);
	int len = strlen(pattern);
	lua_pushinteger(l,(int)scanmem(proc,pattern,len));
	return 1;
}

int alloc(HANDLE proc, int len)
{
	return (int)VirtualAllocEx(proc,NULL,len,MEM_COMMIT,PAGE_EXECUTE_READWRITE);
}

int lua_alloc(lua_State *l)
{
	HANDLE proc = (HANDLE)lua_tointeger(l,1);
	int len = lua_tointeger(l,2);
	lua_pushinteger(l,alloc(proc,len));
	return 1;
}

void injectdll(HANDLE proc, char *dllname)
{
	HANDLE thread;
	void *remotemem;
	DWORD len = strlen(dllname)+1;
	int ret;

	DWORD byteswritten;

	printf("injecting %s...",dllname);
	fflush(stdout);
	remotemem = VirtualAllocEx(proc,NULL,len,MEM_RESERVE|MEM_COMMIT,PAGE_EXECUTE_READWRITE);
	if(!remotemem)
		fprintf(stderr,"Couldn't allocate in thread\n");
	ret = WriteProcessMemory(proc,remotemem,(void *)dllname,len,&byteswritten);
	if(!ret || byteswritten != len)
		fprintf(stderr,"Couldn't write process memory\n");

	HMODULE kernel32 = GetModuleHandle("kernel32.dll");
	LPVOID LoadLib = (void *)GetProcAddress(kernel32,"LoadLibraryA");
	if(!LoadLib)
		fprintf(stderr,"Couldn't GetProcAddress\n");

	DWORD fetch=0;
	DWORD n;
	int i;
	for(i=0;true;i++) {
		ReadProcessMemory(proc,LoadLib,&fetch,sizeof(int),&n);
		if(fetch!=0) {
			printf("%d %x\n",i,fetch);
			break;
		}
		//Most everybody sees the first signature
		//SFilip saw the second one-Rootkit?
		/*if(fetch == 0x8b55ff8b || fetch == 0xdd2717e9) {
			printf("%d\n",i);
			break;
		}*/
		Sleep(1);
	}

	thread = CreateRemoteThread(proc,NULL,0,(LPTHREAD_START_ROUTINE)LoadLib,remotemem,0,NULL);
	if(!thread)
		fprintf(stderr,"Injection failed\n");
	printf("done\n");
	WaitForSingleObject(thread,INFINITE);

	//GetExitCodeThread(thread,&module);
	CloseHandle(thread);
	VirtualFreeEx(proc,remotemem,len,MEM_RELEASE);
}

int lua_injectdll(lua_State *l)
{
	HANDLE proc = (HANDLE)lua_tointeger(l,1);
	char *dllname = (char *)lua_tostring(l,2);
	injectdll(proc,dllname);

	lua_pushinteger(l,0);
	return 1;
}

/*
int lua_resume(lua_State *l)
{
	HANDLE proc = (HANDLE)lua_tointeger(1,1);
	ResumeThread(
}*/

int xargc;
char **xargv;

#include <string>
#include <string.h>

char *getargs(void)
{
	std::string s = "";
	for(int i=1;i<xargc;i++) {
		s+=xargv[i];
		s+=" ";
	}
	return strdup(s.c_str());
}

int lua_getargs(lua_State *l)
{
	lua_pushstring(l,getargs());
	return 1;
}

int regopenkey(char *keyname,char *valuename,char *out)
{
	int ret;
	HKEY key;
	DWORD len=1024;
	try {
		ret = RegOpenKeyEx(HKEY_CURRENT_USER,keyname,0,KEY_QUERY_VALUE,&key);
		if(ret != ERROR_SUCCESS || !key) throw("Could not open registry key");
		ret = RegQueryValueEx(key,valuename,0,0,(LPBYTE)out,&len);
		RegCloseKey(key);
		if(ret != ERROR_SUCCESS) throw("Could not open registry value");
		printf("found %s\\%s : %s\n",keyname,valuename,out);
		return true;
	}catch(const char *error) {
		printf("Attempt to open %s\\%s failed: %s\n",keyname,valuename,error);
/*		char buf[1024];
		sprintf(buf,"Attempt to open key %s\\%s failed: %s\n",keyname,valuename,error);
		MessageBox(NULL,buf,"exehack.exe",MB_ICONERROR);*/
		return false;
	}
}


int lua_regopenkey(lua_State *l)
{
	char *keyname = (char *)lua_tostring(l,1);
	char *value = (char *)lua_tostring(l,2);
	char buf[1024];
	int ret = regopenkey(keyname,value,buf);
	if(ret) lua_pushstring(l,buf);
	else lua_pushinteger(l,0);
	return 1;
}

int exists(char *fname)
{
	FILE *f = fopen(fname,"r");
	if(!f) return false;
	fclose(f);
	return true;
}

int lua_exists(lua_State *l)
{
	char *fname = (char *)lua_tostring(l,1);
	lua_pushboolean(l,exists(fname));
	return 1;
}

int lua_getcwd(lua_State *l)
{
	char path[1024];
	getcwd(path,1024);
	lua_pushstring(l,path);
	return 1;
}



//Written by xttocs
PIMAGE_SECTION_HEADER GetImageSectionHeaders(HMODULE hModule, WORD *count)
{
	if (PIMAGE_DOS_HEADER(hModule)->e_magic != IMAGE_DOS_SIGNATURE)
		return 0;

	PIMAGE_NT_HEADERS pNTHeader = PIMAGE_NT_HEADERS((PBYTE)(hModule) + PIMAGE_DOS_HEADER(hModule)->e_lfanew);
	if (pNTHeader->Signature != IMAGE_NT_SIGNATURE)
		return 0;

	*count = pNTHeader->FileHeader.NumberOfSections;
	return PIMAGE_SECTION_HEADER(pNTHeader + 1);
}
//Written by xttocs
PIMAGE_SECTION_HEADER GetImageSectionHeader(HMODULE hModule, char name[8])
{
	WORD count = 0;
	PIMAGE_SECTION_HEADER pSectionHeader = GetImageSectionHeaders(hModule, &count);

	for (WORD i = 0; i < count; i++)
		if (memcmp(pSectionHeader + i, name, 8) == 0) // Maybe this should be strcmp?
			return pSectionHeader + i;

	return 0;
}

int main(int argc, char **argv)
{
	if(getopt_noarg(argc,argv,"--help") | getopt_noarg(argc,argv,"-h")) {
		puts("ExeHack 0.4a  8/24/2006");
		puts("A lua interface for scripting patches");
		puts("Written by Bryan Steinbach aka PipeDream");
		puts("bsteinbach@berkeley.edu");
		puts("Options:");
		puts("-h --help: Print this message");
		puts("-s <scriptname> --script <scriptname>: Load <scriptname>");
		puts(" if not specified, script.lua is default");
		puts("Options not recognized will be passed via getargs() to script");
		return 0;
	}
	
	char *scriptfile = getopt_arg(argc,argv,"-s");
	if(scriptfile == NULL) scriptfile = getopt_arg(argc,argv,"--script");
	if(scriptfile == NULL) scriptfile = "script.lua";

	xargc = argc;
	xargv = argv;

	lua_State *lua = lua_open();

	luaL_openlibs(lua);

	lua_register(lua,"startexe",lua_startexe);
	lua_register(lua,"nopout",lua_nopout);
	lua_register(lua,"xoreax",lua_xoreax);
	lua_register(lua,"patchbyte",lua_patchbyte);
	lua_register(lua,"patchdword",lua_patchdword);
	lua_register(lua,"patchstr",lua_patchstr);
	lua_register(lua,"alloc",lua_alloc);
	lua_register(lua,"injectdll",lua_injectdll);
	lua_register(lua,"getargs",lua_getargs);
	lua_register(lua,"getregpair",lua_regopenkey);
	lua_register(lua,"exists",lua_exists);
	lua_register(lua,"getcwd",lua_getcwd);
	lua_register(lua,"sleep",lua_sleep);

	int s = luaL_loadfile(lua,scriptfile);
	if (s == 0) {
		s = lua_pcall(lua,0,LUA_MULTRET,0);
	}
	report_errors(lua,s);

	lua_close(lua);



	return 0;
}
