/*
 war3err 1.3
 last modified 2/8/2007
 Detector for JASS error conditions 
 Copyright 2006 Bryan Steinbach

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */


#include <windows.h>
#include <stdio.h>

#include <map>

#include "libbs.h"
#include "section.h"
#include "war3err.h"
#include "opcode.h"
#include "warhook.h"
#include "parseconf.h"

FILE *maxoplog;

MAKETYPE(HPLAYER);
MAKETYPE(HGROUP);
MAKETYPE(HLOC);
MAKETYPE(HUNIT);

typedef BOOL (__stdcall *pSFileLoadFile)(LPCSTR lpszFileName, LPVOID *lplpFileData, LPDWORD lpnFileSize, DWORD nExtraBytesToAllocateAtEnd, LPOVERLAPPED lpOverlapped);
typedef BOOL (__stdcall *pSFileUnloadFile)(LPVOID filedata);

pSFileLoadFile SFileLoadFile = NULL;
pSFileUnloadFile SFileUnloadFile = NULL;

typedef int (__stdcall *pjExecuteFunc)(NATIVESTR *);

pjExecuteFunc war3_jExecuteFunc;


typedef int (__stdcall *pjRemoveLocation)(HLOC);
pjRemoveLocation war3_jRemoveLocation;


typedef HGROUP (__stdcall *pjCreateGroup)(void);
typedef int (__stdcall *pjDestroyGroup)(HGROUP);

pjCreateGroup war3_jCreateGroup;
pjDestroyGroup war3_jDestroyGroup;

typedef HPLAYER (__stdcall *pjPlayer)(int);
typedef HPLAYER (__stdcall *pjGetLocalPlayer)();
typedef void (__stdcall *pjDisplayTimedTextToPlayer)(HPLAYER p,float *x, float *y, float *t, NATIVESTR *s);

pjPlayer war3_jPlayer;

pjPlayer jPlayer;
pjGetLocalPlayer jGetLocalPlayer;
pjDisplayTimedTextToPlayer jDisplayTimedTextToPlayer;

void initjnative()
{
	jPlayer = (pjPlayer)GetNativeAddress("Player");
	if(jPlayer == NULL) {
		dprintf("Couldn't find native Player\n");
		return;
	}
	dprintf("Player: %x\n",jPlayer);
	jDisplayTimedTextToPlayer = (pjDisplayTimedTextToPlayer)GetNativeAddress("DisplayTimedTextToPlayer");
	if(jDisplayTimedTextToPlayer == NULL) {
		dprintf("Couldn't find native DisplayTimedTextToPlayer\n");
		return;
	}
	dprintf("DisplayTimedTextToPlayer: %x\n",jDisplayTimedTextToPlayer);
	jGetLocalPlayer = (pjGetLocalPlayer)GetNativeAddress("GetLocalPlayer");
	if(jGetLocalPlayer == NULL) {
		dprintf("Couldn't find native GetLocalPlayer\n");
		return;
	}
	dprintf("GetLocalPlayer: %x\n",jGetLocalPlayer);
}

/*
 * sends error message to Player(0)
 */
void GamePrint(VM *vm, char *s)
{
	static int needinit = true;
	if(needinit) { initjnative(); needinit = false; }

	NATIVESTR *c = vm->NativeStrMap(s);
	HPLAYER p = jGetLocalPlayer();
	float x = 0.0;
	float y = 0.0;
	float t = 120.0;
	jDisplayTimedTextToPlayer(p,&x,&y,&t,c);
}

void GamePrintf(VM *vm, char *format, ...)
{
	char buf[8192];
	va_list args;
	va_start(args,format);
	vsprintf(buf,format,args);
	GamePrint(vm,buf);
}

void ErrPrintf(VM *vm, char *format, ...)
{
	char buf1[8192];
	char buf2[8192];
	va_list args;
	va_start(args,format);
	vsprintf(buf1,format,args);
	sprintf(buf2,"|cffff0000%s|r",buf1);
	GamePrint(vm,buf2);
	dprintf("%s\n",buf1);
}

void patchbytes(void *dest, void *src, int len)
{
	DWORD old;
	DWORD old2;

	if(!VirtualProtect(dest,len,PAGE_READWRITE,&old)) throw("couldn't set PAGE_READWRITE on page");
	memmove(dest,src,len);
	if(!VirtualProtect(dest,len,old,&old2)) throw("couldn't restore permissions on page");
}

void HookNative(Section &code,Section &data,unsigned char *nativename, int nativenamelen, unsigned char *regpat,unsigned char *regmask, int regpatlen, long *war3_native,long *hooknative)
{
	dprintf("nativename: %s\n",nativename);
	dprintf("hooknative: %x\n", hooknative);
	dprintf("hooknativedata: %x\n", *hooknative);
	unsigned char *nativenameaddr = data.Search((unsigned char *)nativename,NULL,nativenamelen);
	dprintf("nativenameaddr: %x\n",nativenameaddr);
	*(unsigned char **)&regpat[1] = nativenameaddr;
	unsigned char *regaddr = code.Search(regpat,regmask,regpatlen);
	dprintf("regaddr: %x\n",regaddr);
	*war3_native = *(long *)&regaddr[6];
	dprintf("war3_native: %x\n",*war3_native);
	patchbytes((void *)&regaddr[6],(void *)&hooknative,sizeof(long));
	dprintf("\n");
}

VM *curvm;
opcode *curop;


extern pNativeStrGet NativeStrGet;

int countargs(opcode *func)
{
	int i;
	for(i=0;func[i].optype == OP_POPFUNCARG;i++) {}
	return i;
}

int __stdcall ExecuteFunc(NATIVESTR *n)
{
	char *s = NativeStrGet(n);
	opcode *op = (opcode *)HTLookup(curvm->functable,s);
	if(HTLookup(curvm->functable,s) == NULL) {
		ErrPrintf(curvm,"Error - ExecuteFunc non existent \"%s\" in \"%s\"",s,getfuncname(curvm,curop));
		return 0;
	}

	int narg = countargs(op);
	if(narg > 0) {
		ErrPrintf(curvm,"Error - ExecuteFunc \"%s\" with %d instead of 0 arguments in \"%s\"",s,narg,getfuncname(curvm,curop));
		return 0;
	}
	return war3_jExecuteFunc(n);

}

void InstallExecuteFuncHook(Section &code, Section &data)
{

	HookNative(code,data,ExecuteFuncStrPattern,sizeof(ExecuteFuncStrPattern),NativeRegPattern,NativeRegMask,sizeof(NativeRegPattern),(long *)&war3_jExecuteFunc,(long *)ExecuteFunc);
}

class LeakTrace {
	public:
		LeakTrace(char *name);
		void Alloc(int x);
		void Free(int x);
		void Report(void);
	private:
		char *mName;
		int mNcreated;
		int mNactive;
		std::map<char *,int,ltstr> mFunc_allocs;
		std::map<int,char *> mAlloced_by;
};

LeakTrace::LeakTrace(char *name) : mName(name),mNcreated(0),mNactive(0) {}
void LeakTrace::Alloc(int x)
{
	mNcreated++;
	mNactive++;

	char *func;
	func = getfuncname(curvm,curop);
	mAlloced_by[x] = func;
	if(mFunc_allocs.find(func) == mFunc_allocs.end()) {
		mFunc_allocs[func] = 1;
	} else {
		mFunc_allocs[func] = mFunc_allocs[func] + 1;
	}	
}
typedef struct {
	char *name;
	int n;
} pair;
int cmppair(const void *x,const void *y)
{
	pair *a = (pair *)x;
	pair *b = (pair *)y;
	if(a->n < b->n) return 1;
	else if(a->n == b->n) return 0;
	else return -1;
}

void LeakTrace::Report()
{
	ErrPrintf(curvm,"Total %ss created: %d",mName,mNcreated);
	ErrPrintf(curvm,"Total undestroyed %ss: %d",mName,mNactive);
	
	std::map<char *,int,ltstr>::iterator x;
	std::map<char *,int,ltstr>::iterator end = mFunc_allocs.end();

	int nf = mFunc_allocs.size();

	pair p[nf];

	int i=0;
	for(x = mFunc_allocs.begin();x!=end;x++,i++) {
		p[i].name = (*x).first;
		p[i].n = (*x).second;
	}
	
	qsort(p,nf,sizeof(pair),cmppair);

	int nprint = 5;
	if(nf < 5) {
		nprint = nf;
	}
	for(i=0;i<nprint;i++) {
		if(p[i].n > 3)
			ErrPrintf(curvm,"%s with %d active %ss",p[i].name,p[i].n,mName);
	}
	ErrPrintf(curvm,"");
}

void LeakTrace::Free(int x)
{
	char *func;
	if(mAlloced_by.find(x) == mAlloced_by.end()) {
		ErrPrintf(curvm,"Unknown %s freed in \"%s\"",mName,getfuncname(curvm,curop));
	}
	else {
		func = mAlloced_by[x];
		if(func == NULL) {
			ErrPrintf(curvm,"double free of %s in \"%s\"",mName,getfuncname(curvm,curop));
		} else {
			mNactive--;
			mFunc_allocs[func] = mFunc_allocs[func] - 1;
			mAlloced_by[x] = NULL;
		}
	}
}

LeakTrace *loctrace;

int __stdcall RemoveLocation(HLOC l)
{
	if(l.z == 0) {
		loctrace->Report();
		return 0;
	} else {
		loctrace->Free(l.z);
		return war3_jRemoveLocation(l);
	}
}

extern "C" {
	void LocationHook();
	void LocationPatch();
	void LocationPatchEnd();
	unsigned char *LocationPatchdest;
	long LocationPatchsize;

	void __cdecl LocationHandler(int x)
	{
		loctrace->Alloc(x);
	}
}

void InstallLocTracePatch(Section &code,Section &data)
{
	loctrace = new LeakTrace("location");

	LocationPatchsize = (long)LocationPatchEnd-(long)LocationPatch;
	LocationPatchdest = code.Search(LocationPatchPattern,LocationPatchMask,sizeof(LocationPatchPattern));
	dprintf("Location: %x\n",LocationPatchdest);
	patchbytes((void *)LocationPatchdest,(void *)LocationPatch,LocationPatchsize);

	HookNative(code,data,RemoveLocationStrPattern,sizeof(RemoveLocationStrPattern),NativeRegPattern,NativeRegMask,sizeof(NativeRegPattern),(long *)&war3_jRemoveLocation,(long *)RemoveLocation);
}


LeakTrace *grouptrace;

HGROUP __stdcall CreateGroup(void)
{
	HGROUP g= war3_jCreateGroup();
	grouptrace->Alloc(g.z);
	return g;
}

int __stdcall DestroyGroup(HGROUP g)
{
	if(g.z == 0) {
		grouptrace->Report();
		return 0;
	} else {
		grouptrace->Free(g.z);
		return war3_jDestroyGroup(g);
	}
}


void InstallGroupTracePatch(Section &code,Section &data)
{
	HookNative(code,data,CreateGroupStrPattern,sizeof(CreateGroupStrPattern),NativeRegPattern,NativeRegMask,sizeof(NativeRegPattern),(long *)&war3_jCreateGroup,(long *)CreateGroup);
	HookNative(code,data,DestroyGroupStrPattern,sizeof(DestroyGroupStrPattern),NativeRegPattern,NativeRegMask,sizeof(NativeRegPattern),(long *)&war3_jDestroyGroup,(long *)DestroyGroup);
	grouptrace = new LeakTrace("group");
}

HPLAYER __stdcall safe_Player(int p)
{
	if((p<0) || (p>15)) {
		char *func = getfuncname(curvm,curop);
		if(func == NULL) {
			dprintf("NULL in safe_Player\n");
		} else {
			ErrPrintf(curvm,"Invalid \"Player(%d)\" in function \"%s\"",p,func);
		}
		return war3_jPlayer(0);
	}
	return war3_jPlayer(p);
}

void InstallPlayerPatch(Section &code, Section &data)
{
	HookNative(code,data,JassPlayerStrPattern,sizeof(JassPlayerStrPattern),NativeRegPattern,NativeRegMask,sizeof(NativeRegPattern),(long *)&war3_jPlayer,(long *)safe_Player);
}

typedef int (__stdcall *pSetUnitAbilityLevel)(int, int, int);
pSetUnitAbilityLevel war3_jSetUnitAbilityLevel;

int __stdcall SetUnitAbilityLevel(int unit, int abil, int level)
{
	char *s = (char *)&abil;
	dprintf("call SetUnitAbilityLevel(%x,'%c%c%c%c',%d) in ",unit,s[3],s[2],s[1],s[0],level);
	dprintf("function \"%s\"\n",getfuncname(curvm,curop));
	return war3_jSetUnitAbilityLevel(unit,abil,level);
}

void InstallSUALPatch(Section &code, Section &data)
{
	HookNative(code,data,JassSetUnitAbilityLevelStrPattern,sizeof(JassSetUnitAbilityLevelStrPattern),NativeRegPattern,NativeRegMask,sizeof(NativeRegPattern),(long *)&war3_jSetUnitAbilityLevel,(long *)SetUnitAbilityLevel);
}



int gamedllhacked = false;
char *regnat_backup[32];

extern "C" {
	void RegWaitHook();
	void RegWaitPatch();
	void RegWaitPatchEnd();
	long RegWaitPatchsize = (long)RegWaitPatchEnd-(long)RegWaitPatch;
	unsigned char *pRegWaitPatchdest;

	void __cdecl RegWaitHandler()
	{
		int i;
		for(i=0;!gamedllhacked;i++) { Sleep(1); }
		dprintf("Waited for %d before letting RegisterNatives run\n",i);

		//restore overwritten chunk of registernatives
		patchbytes((void *)pRegWaitPatchdest,(void *)regnat_backup,RegWaitPatchsize);
	}
}

void InstallRegWaitPatch(Section &code)
{
	pRegWaitPatchdest = REGISTERNATIVESOFFSET+code.Search(RegisterNativesPattern,RegisterNativesMask,sizeof(RegisterNativesPattern));
	if(!pRegWaitPatchdest) throw("RegisterNatives signature not found");
	dprintf("Register natives signature found at: %x\n",pRegWaitPatchdest);
	memcpy(regnat_backup,pRegWaitPatchdest,RegWaitPatchsize);
	patchbytes((void *)pRegWaitPatchdest,(void *)RegWaitPatch,RegWaitPatchsize);
}


class Profiler {
	public:
		Profiler(FILE *fout);
		void Log(char *s);
		void WriteLog();
	protected:
		FILE *mFout;
		std::map<char *,int, ltstr> mLog;
};

Profiler::Profiler(FILE *fout) : mFout(fout) {}

void Profiler::Log(char *s)
{
	if(mLog.find(s) == mLog.end()) mLog[s] = 1;
	else mLog[s] = mLog[s]+1;
}

void Profiler::WriteLog()
{
	rewind(mFout);
	pair pairs[mLog.size()];
	std::map<char *,int,ltstr>::iterator p;
	std::map<char *,int,ltstr>::iterator end = mLog.end();

	unsigned int i=0;
	for(p=mLog.begin();p!=end;p++,i++) {
		pairs[i].name = (*p).first;
		pairs[i].n = (*p).second;
	}

	qsort(pairs,mLog.size(),sizeof(pair),cmppair);

	fprintf(mFout,"%d\n",mLog.size());

	for(i=0;i<mLog.size();i++) {
		fprintf(mFout,"%s %d\n",pairs[i].name,pairs[i].n);
	}
	fflush(mFout);
}

Profiler *funclog=NULL,*varlog=NULL;

void InitProfile()
{
	FILE *funcout = fopen("jasscalllog.txt","w");
	if(!funcout)
		dprintf("Couldn't open jasscallog.txt");
	else
		funclog = new Profiler(funcout);
	FILE *varout = fopen("varuselog.txt","w");
	if(!varout)
		dprintf("Couldn't open varuselog.txt");
	else
		varlog = new Profiler(varout);
}

void logfunction(int sindex)
{
	if(funclog == NULL) return;
	char *s = curvm->symtable->Get(sindex);
	if(strcmp("LogJASSCalls",s)==0)
		funclog->WriteLog();
	else
		funclog->Log(s);
}

void logvar(int sindex)
{
	if(varlog == NULL) return;
	char *s = curvm->symtable->Get(sindex);
	if(strcmp("LogVariableUse",s)==0)
		varlog->WriteLog();
	else
		varlog->Log(s);
}

char *op_unknown = "unknown";

char *optypenames[] = 
{
	op_unknown,
	op_unknown,
	"jump",
	"func_start",
	"func_end",
	"local",
	"global",
	"constant",
	"pop",
	op_unknown,
	op_unknown, // 10
	"clean stack",
	"literal",
	"retval",
	"getvar",
	"code",
	"getvar[]",
	"setvar",
	"setvar[]",
	"push",
	"setreg",
	"callnative",
	"callfunc",
	"typecast",
	op_unknown,
	op_unknown,
	"==",
	op_unknown,
	"<=",
	"<",
	">=",
	">",
	"add",
	"subtract",
	"multiply",
	"divide",
	op_unknown,
	"negate",
	"bnot",
	"return",
	"jump dest",
	"jump if true",
	"jump if false",
	op_unknown,
	op_unknown,
	op_unknown
};

char *TYPENAME[] = {
	"unknown","unknown","null","code","integer","real","string","handle","boolean"
};

bool tracebytecode = false;
FILE *bytecodelog = NULL;


/*
 * Cache this information through range sets
 */
char *getfn(opcode *op)
{
	int i;
	for(i=0;op[i].optype != OP_FUNCTION;i--);
	return curvm->symtable->Get(op[i].arg);
}

void logvtype(int x)		//needs bounds checking
{
	if(x > 8) {
		fprintf(bytecodelog,":%s array",TYPENAME[x-8]);
	} else {
		fprintf(bytecodelog,":%s",TYPENAME[x]);
	}
}

void logbc(opcode *op)
{
	fprintf(bytecodelog,"%s :: ",getfn(op));
	fprintf(bytecodelog,"%d",op->r1);
	switch(op->optype) {
		case OP_GETARRAY:
			logvtype(op->r1);
	}
	fprintf(bytecodelog," ");
	fprintf(bytecodelog,"%d",op->r2);
	switch(op->optype) {
		case OP_LITERAL:
		case OP_GETVAR:
		case OP_CODE:
			logvtype(op->r2);
	}
	fprintf(bytecodelog," ");
	fprintf(bytecodelog,"%d",op->r3);
	switch(op->optype) {
		case OP_LOCAL:
		case OP_GLOBAL:
		case OP_CONSTANT:
		case OP_POPFUNCARG:
			logvtype(op->r3);
	}
	fprintf(bytecodelog," ");

	fprintf(bytecodelog,"0x%x:%s ",op->optype,optypenames[op->optype]);
	fprintf(bytecodelog,"%d",op->arg);
	switch(op->optype) {
		case OP_LITERAL:		//if it's J_STRING, follow through
			if (op->r2 != J_STRING) break;
		case OP_NATIVE:
		case OP_JASSCALL:
		case OP_POPFUNCARG:
		case OP_LOCAL:
		case OP_GLOBAL:
		case OP_CONSTANT:
		case OP_SETVAR:
		case OP_GETVAR:
		case OP_SETARRAY:
		case OP_GETARRAY:
			fprintf(bytecodelog,":%s", curvm->symtable->Get(op->arg));
			break;
	}
	fprintf(bytecodelog,"\n");
	switch(op->optype) {
		case OP_SETARRAY:
		case OP_SETVAR:
			if(op->r2 != J_HANDLE) break;
		case OP_NATIVE:
			fflush(bytecodelog);

	}
}

//Tracks opcount
extern "C" {
	void OpHook();
	void OpPatch();
	void OpPatchEnd();

	void __cdecl OpHandler(opcode *ops, VM *vm)
	{
		curvm = vm;
		curop = ops;
		if(tracebytecode)
			logbc(ops);

		switch(ops->optype) {
			case OP_JASSCALL:
				logfunction(ops->arg);
				break;
			case OP_LOCAL:
			case OP_GETVAR:
			case OP_SETVAR:
				logvar(ops->arg);
				break;
		}

	}
}
void InstallOpPatch(Section &code)
{
	long OpPatchsize = (long)OpPatchEnd-(long)OpPatch;
	unsigned char *pOpPatchdest = code.Search(JassOpPatchPattern,NULL,sizeof(JassOpPatchPattern));
	if(!pOpPatchdest) throw("Op signature not found");
	dprintf("Op signature found at: %x\n",pOpPatchdest);
	patchbytes((void *)pOpPatchdest,(void *)OpPatch,OpPatchsize);
}

extern "C" {
	void OpLimitHook(void);
	void OpLimitPatch();
	void OpLimitPatchEnd();

	void __cdecl CountHandler(int max, int count, opcode *ops, VM *vm)
	{
		static int highest = 0;
		if(count > highest) {
			highest = count;
			fseek(maxoplog,0,SEEK_SET);
			fprintf(maxoplog,"%d\n",count);
			fflush(maxoplog);
		}
		if(count == max) {
			char *func = getfuncname(vm,ops);
			if(func == NULL) dprintf("null in counthandler\n");
			else {
				ErrPrintf(vm,"Hit op limit in function %s",func);
			}
		}
	}
}
void InstallOpLimitPatch(Section &code)
{
	long OpLimitPatchsize = (long)OpLimitPatchEnd-(long)OpLimitPatch;

	unsigned char *pOpLimitPatchdest;
	pOpLimitPatchdest = code.Search(JassOpLimitPatchPattern,JassOpLimitPatchMask,sizeof(JassOpLimitPatchPattern));
	if(!pOpLimitPatchdest) throw("OpLimit signature not found");
	
	dprintf("OpLimit signature found at: %x\n",pOpLimitPatchdest);
	dprintf("%x %x\n",OpLimitPatchEnd,OpLimitPatch);
	patchbytes((void *)pOpLimitPatchdest,(void *)OpLimitPatch,OpLimitPatchsize);
}

extern "C" {
	void DivZeroHook(void);
	void DivZeroPatch();
	void DivZeroPatchEnd();

	void __cdecl DivZeroHandler(opcode *ops,VM *vm)
	{

		char *func = getfuncname(vm,ops);
		if(func == NULL) dprintf("null in DivZeroHandler\n");
		else {
			ErrPrintf(vm,"Divide by zero error in function %s",func);
		}
	}
}
void InstallDivZeroPatch(Section &code)
{
	long DivZeroPatchsize = (long)DivZeroPatchEnd-(long)DivZeroPatch;

	unsigned char *pDivZeroPatchdest;
	pDivZeroPatchdest = JassDivZeroPatchOffset + code.Search(JassDivZeroPatchPattern,JassDivZeroPatchMask,sizeof(JassDivZeroPatchPattern));
	if(!pDivZeroPatchdest) throw("DivZero signature not found");
	dprintf("DivZero signature found at: %x\n",pDivZeroPatchdest);
	patchbytes((void *)pDivZeroPatchdest,(void *)DivZeroPatch,DivZeroPatchsize);
}

extern "C" void PausePatch();
extern "C" void PausePatchEnd();
void InstallNoPausePatch(Section &code)
{
	long PausePatchsize = (long)PausePatchEnd - (long)PausePatch;
	try {
		unsigned char *target = code.Search(PausePatchPattern,PausePatchMask,sizeof(PausePatchPattern));
		if(!target) throw("Pause patch signature not found");
		dprintf("Pause patch signature found at: %x\n",target);
		patchbytes((void *)target,(void *)PausePatch,PausePatchsize);
	} catch(const char *s) {
		dprintf("InstallNoPausePatch error: %s\n",s);
	}
}


extern "C" {

	void VarUseHook(void);
	void VarUsePatch();
	void VarUsePatchEnd();
	
	void __cdecl VarUseHandler(opcode *ops,VM *vm)
	{
		char *func = getfuncname(vm,ops);
		if(func == NULL) dprintf("null in VarUseHandler\n");
		else {
			//ErrPrintf(vm,"Uninitialized variable \"%s\" used in function \"%s\"",vm->symtable->Get(ops->arg),func);
			ErrPrintf(curvm,"Uninitialized variable \"%s\" used in function \"%s\"",curvm->symtable->Get(curop->arg),func);
		/*	ops--;
			ErrPrintf(vm,"opcode: %2x %2x %2x %2x:%s %x",ops->r1,ops->r2,ops->r3,ops->optype,optypenames[ops->optype],ops->arg);
			ops++;
			ErrPrintf(vm,"opcode: %2x %2x %2x %2x:%s %x",ops->r1,ops->r2,ops->r3,ops->optype,optypenames[ops->optype],ops->arg);
			ops++;
			ErrPrintf(vm,"opcode: %2x %2x %2x %2x:%s %x",ops->r1,ops->r2,ops->r3,ops->optype,optypenames[ops->optype],ops->arg);*/
		}
	}
}
void InstallVarUsePatch(Section &code)
{
	long VarUsePatchsize = (long)VarUsePatchEnd - (long)VarUsePatch;

	unsigned char *pVarUsePatchdest;
	pVarUsePatchdest = JassVarUsePatchOffset + code.Search(JassVarUsePatchPattern,JassVarUsePatchMask,sizeof(JassVarUsePatchPattern));
	if(!pVarUsePatchdest) throw("VarUse signature not found");
	dprintf("VarUse signature found at: %x\n",pVarUsePatchdest);
	patchbytes((void *)pVarUsePatchdest,(void *)VarUsePatch,VarUsePatchsize);
}

ParseConf *conf;

DWORD __stdcall OnGameLoad(LPVOID param)
{
	HMODULE hGame = NULL;
	while(1) {
		hGame = GetModuleHandle("game.dll");
		if(hGame != NULL) break;
		Sleep(1);
	}
	try {
		Section gamecode(GetModuleHandle("game.dll"),".text\0\0\0");
		InstallRegWaitPatch(gamecode);

		Section gamedata(GetModuleHandle("game.dll"),".data\0\0\0");

		initwarhook(gamecode);

		InstallOpLimitPatch(gamecode);
		InstallDivZeroPatch(gamecode);
		InstallVarUsePatch(gamecode);
		InstallOpPatch(gamecode);
		InstallPlayerPatch(gamecode,gamedata);
		InstallGroupTracePatch(gamecode,gamedata);
		InstallLocTracePatch(gamecode,gamedata);
		InstallExecuteFuncHook(gamecode,gamedata);
		if(conf->VarOn("nopause"))
			InstallNoPausePatch(gamecode);
		if(conf->VarOn("bytecodetrace"))
			tracebytecode = true;

		dprintf("game.dll patched successfully\n");
		gamedllhacked = true;	//allow register natives to proceed
	} catch(const char *s) {
		dprintf("Error: %s\n",s);
		char buf[1024];
		sprintf(buf,"Error: %s\n",s);
		MessageBox(NULL,buf,"War3Err.dll",MB_ICONERROR);
	}
	return 0;
}

extern "C" BOOL __stdcall DllMain(HMODULE module, DWORD reason, LPVOID reserved)
{
	if(reason == DLL_PROCESS_ATTACH) {
		DisableThreadLibraryCalls(module);
		initlibbs("war3err.txt");
		maxoplog = fopen("maxoplog.txt","w");
		conf = new ParseConf("war3err.conf");
		//bytecodelog = fopen("bytecode.txt","w");
		bytecodelog = popen("logd.exe bytecode.txt","w");
		InitProfile();
		CloseHandle(CreateThread(NULL,16384,OnGameLoad,NULL,0,NULL));
	}
	return true;
}
