/*
Structures, addresses and signatures of Warcraft
*/

#pragma once

#include "opcode.h"

typedef struct {
	char unk[0x14];
	long vartype;
	long vartype2;
	long value;
} variable;


typedef struct {
	char x[0x14];
	char *str;
} jstr;

typedef struct {
	long a;
	long b;
	jstr **js;
} sym1;

/*
 * mov ecx,[ecx+8]	//a = symtable->s1
 * mov edx,[ecx+8]	//b = symtable->s1->s2
 * mov eax,[edx+eax*4]	//c = (symtable->s1->s2).s3[x]
 * mov eax,[eax+0x14]	//d = (symtable->s1->s2).s3[x].str
 * return eax
 * Get:
 * Symtable->s1->js[x]->str
 */

class Symtable {
	public:
		long a;
		long b;
		sym1 *s1;
		char *Get(int x);
		int Map(char *s);
};

/*
 * Use this hash table to look up function name -> opcode start
 * probably works for other stuff
 */

struct HT_Bucket;
struct Hashtable;

/*
 * this node is specific for char * -> void * map
 */


typedef struct HT_Node {
	long hash;
	struct HT_Bucket *parent;
	HT_Node *next;
	char unk[0x8];
	char *key;
	void *value;
} HT_Node;

/*
 * this bucket is meant for generic nodes
 * offset +4 is how to crawl the node list
 * this has been hardcoded into HTLookup for now
 */

typedef struct HT_Bucket {	//size = 12
	long offset;
	long unk;
	HT_Node *list;
} HT_Bucket;

typedef struct Hashtable {
	char unk1[0x1C];
	HT_Bucket *buckets;	//0x1C
	long unkl;		//0x20
	long bitmask;		//0x24 //determines number of elements
} Hashtable;

//size should be a power of two
Hashtable *HTCreate(int size);
void *HTLookup(Hashtable *HT, char *key);
void HTInsert(Hashtable *HT, char *key, void *value);
void HTRemove(Hashtable *HT, char *key);


#define MAKETYPE(JTYPE) typedef struct { long z; } JTYPE

MAKETYPE(CACHESTR);

typedef struct {
	char unk[16];
} NATIVESTR;

typedef struct {
	long a;
	long b;
	NATIVESTR *ns;
} NSTABLE;

/*
 * binary ops:
 * r1 is index for t on lhs
 * r2 is index for t on rhs
 * r3 is index of destination
 *
 * game_0x6F6AFF00 assigns vartype = vartype2
 */

typedef struct {
	long unk1;
	char unk2[0x10];
	long unk3;	//0x14
	long vartype;	//0x18 up
	long vartype2;	//0x1C
	long value;	//0x20
	long d;
} Temp;

class VM {
	public:
		char unk1[0x20];
		opcode *curop;
		char unk2[0x28];	//t starts at 0x4C
		Temp t[256];
		char unk3[0x8];
		Symtable *symtable; // 0x2854
		long a;		//0x2858
		long b;		//0x285C
		long c;		//0x2860
		long d;	// 0x2864 this one has something to do with vars
		long e;		//0x2868
		long f;		//0x286C
		NSTABLE *nstable;		//0x2870 used for cache_to_nstr
		char unk5[0x0C];	//0x2874
		Hashtable *functable; 	//0x2880
		char unk[0x18];
		long DecrementHandleFunc;	//0x289C
		long HandleTable;	//0x28A0
		
		long unktable1;
		long unktable2;
		long unktable3; //opcode address -->> int  (19?)

		int Execute(opcode *funcstart, int unk1, int maxops, int unk2);
		NATIVESTR *NativeStrMap(char *s);

	private:
		NATIVESTR *cache_to_nstr(CACHESTR s);
};

typedef union {
	int (VM::*Execute)(opcode *funcstart, int unk1, int maxops, int unk2);
	long address;
} VM_Execute_Cast;

typedef struct {
	int nargc;
	TYPES rt;
} finfo;

typedef struct {
	char unk[0x18];
	char *name;
	int address;
	int narg;
	char *prototype;
} native_info;




void *GetNativeAddress(char *name);
char *getfuncname(VM *vm, opcode *ops);

class Section;
void initwarhook(Section &code);

typedef CACHESTR 	(__fastcall *pNativeStrMap)(char *str);	//same thing as char_to_cache ?
typedef char *			(__fastcall *pNativeStrGet)(NATIVESTR *s);



